# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import re

import typepy

import pyparsing as pp

from ._common import _to_unicode


class TcQdiscParser(object):

    def __init__(self):
        self.__clear()

    def parse(self, text):
        if typepy.is_null_string(text):
            raise ValueError("empty text")

        text = text.strip()

        for line in text.splitlines():
            if typepy.is_null_string(line):
                continue

            line = _to_unicode(line.lstrip())

            if re.search("qdisc netem|qdisc tbf", line) is None:
                continue

            self.__clear()

            if re.search("qdisc netem", line) is not None:
                self.__parse_netem_param(line, "parent", pp.hexnums + ":")

            self.__parse_netem_param(line, "delay", pp.nums + ".")
            self.__parse_netem_delay_distro(line)
            self.__parse_netem_param(line, "loss", pp.nums + ".")
            self.__parse_netem_param(line, "duplicate", pp.nums + ".")
            self.__parse_netem_param(line, "corrupt", pp.nums + ".")
            self.__parse_netem_param(line, "reorder", pp.nums + ".")
            self.__parse_tbf_rate(line)

            yield self.__parsed_param

    def __clear(self):
        self.__parsed_param = {}

    def __parse_netem_delay_distro(self, line):
        parse_param_name = "delay"
        pattern = (
            pp.SkipTo(parse_param_name, include=True) +
            pp.Word(pp.nums + ".") + pp.Literal("ms") +
            pp.Word(pp.nums + ".") + pp.Literal("ms"))

        try:
            parsed_list = pattern.parseString(line)
            self.__parsed_param[parse_param_name] = parsed_list[2]
            self.__parsed_param["delay-distro"] = parsed_list[4]
        except pp.ParseException:
            pass

    def __parse_netem_param(self, line, parse_param_name, word_pattern):
        pattern = (
            pp.SkipTo(parse_param_name, include=True) +
            pp.Word(word_pattern))

        try:
            result = pattern.parseString(_to_unicode(line))[-1]
            if typepy.is_not_null_string(result):
                self.__parsed_param[parse_param_name] = result
        except pp.ParseException:
            pass

    def __parse_tbf_rate(self, line):
        parse_param_name = "rate"
        pattern = (
            pp.SkipTo(parse_param_name, include=True) +
            pp.Word(pp.alphanums + "." + ":"))

        try:
            result = pattern.parseString(line)[-1]
            if typepy.is_not_null_string(result):
                result = result.rstrip("bit")
                self.__parsed_param[parse_param_name] = result
        except pp.ParseException:
            pass



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import


class NetworkInterfaceNotFoundError(Exception):
    """
    Exception raised when network interface not found.
    """


class ModuleNotFoundError(Exception):
    """
    Exception raised when mandatory kernel module not found.
    """


class TcCommandExecutionError(Exception):
    """
    Exception raised when failed to execute a ``tc`` command.
    """


class TcAlreadyExist(TcCommandExecutionError):
    """
    Exception raised when a traffic shaping rule already exist.
    """


class EmptyParameterError(ValueError):
    """
    Exception raised when a parameter value is empty value.
    """


class InvalidParameterError(ValueError):
    """
    Exception raised when an invalid parameter specified for
    a traffic shaping rule.
    """


class UnitNotFoundError(InvalidParameterError):
    """
    """



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from collections import namedtuple
import re

import typepy

from ._error import (
    InvalidParameterError,
    UnitNotFoundError,
)
from ._logger import logger

ByteUnit = namedtuple("ByteUnit", "regexp factor")


class Humanreadable(object):
    __UNIT_LIST = [
        ByteUnit(regexp=re.compile("^b$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^bps$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^k$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^kbps$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^m$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^mbps$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^g$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^gbps$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^t$", re.IGNORECASE), factor=4),
        ByteUnit(regexp=re.compile("^gbps$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^p$", re.IGNORECASE), factor=5),
    ]
    __RE_NUMBER = re.compile("^[\-\+]?[0-9\.]+")

    def __init__(self, readable_size, kilo_size=1024):
        """
        String converter that humanreadable byte size to a number.

        :param str readable_size: human readable size (bytes). e.g. 256 M
        :param int kilo_size: size of ``kilo``. 1024 or 1000
        """

        self.__readable_size = readable_size
        self.kilo_size = kilo_size

        self.__validate_kilo_size()

    def to_bit(self):
        """
        :raises ValueError:
        """

        logger.debug("readable_size: {}".format(self.__readable_size))

        if not typepy.is_not_null_string(self.__readable_size):
            raise TypeError(
                "readable_size must be a string: actual={}".format(
                    self.__readable_size))

        self.__readable_size = self.__readable_size.strip()

        try:
            size = self.__RE_NUMBER.search(self.__readable_size).group()
        except AttributeError:
            raise InvalidParameterError(
                "invalid value: {}".format(self.__readable_size))
        size = float(size)
        if size < 0:
            raise InvalidParameterError(
                "size must be greater or equals to zero")

        unit = self.__RE_NUMBER.sub("", self.__readable_size).strip().lower()

        return size * self.__get_coefficient(unit)

    def to_kilo_bit(self):
        """
        :param str readable_size: human readable size (bytes). e.g. 256 M
        :raises ValueError:
        """

        return self.to_bit() / self.kilo_size

    def __validate_kilo_size(self):
        if self.kilo_size not in [1000, 1024]:
            raise ValueError("invalid kilo size: {}".format(self.kilo_size))

    def __get_coefficient(self, unit_str):
        self.__validate_kilo_size()

        for unit in self.__UNIT_LIST:
            if unit.regexp.search(unit_str):
                return self.kilo_size ** unit.factor

        raise UnitNotFoundError("unit not found: value={}".format(unit_str))



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

from subprocrunner import SubprocessRunner
from typepy.type import RealNumber

from tcconfig._const import TcCommand
from tcconfig._converter import Humanreadable


def is_invalid_param(
        rate, delay, packet_loss, packet_duplicate, corrupt, reordering):
    param_value_list = [
        delay,
        packet_loss,
        packet_duplicate,
        corrupt,
        reordering,
    ]

    print(param_value_list)

    is_invalid = all([
        not RealNumber(param_value).is_type() or param_value <= 0
        for param_value in param_value_list
    ])

    try:
        Humanreadable(rate, kilo_size=1000).to_bit()
    except (TypeError, ValueError):
        pass
    else:
        is_invalid = False

    return is_invalid


def execute_tcdel(device):
    SubprocessRunner(
        "{:s} --device {}".format(TcCommand.TCDEL, device),
        dry_run=False).run()



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import errno
import json
import sys

import logbook
import subprocrunner

from ._argparse_wrapper import ArgparseWrapper
from ._common import (
    verify_network_interface,
    write_tc_script,
)
from ._const import (
    VERSION,
    TcCommand,
    TcCommandOutput,
)
from ._error import NetworkInterfaceNotFoundError
from ._logger import (
    LOG_FORMAT_STRING,
    logger,
    set_log_level,
)
from .parser.shaping_rule import TcShapingRuleParser


logbook.StderrHandler(
    level=logbook.DEBUG, format_string=LOG_FORMAT_STRING).push_application()


def parse_option():
    parser = ArgparseWrapper(VERSION)

    group = parser.parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--device", action="append", required=True,
        help="network device name (e.g. eth0)")
    group.add_argument(
        "--ipv6", dest="ip_version", action="store_const",
        const=6, default=4,
        help="""
        Display IPv6 shaping rules.
        Defaults to show IPv4 shaping rules.
        """)

    return parser.parser.parse_args()


def main():
    options = parse_option()

    set_log_level(options.log_level)

    try:
        subprocrunner.Which("tc").verify()
    except subprocrunner.CommandNotFoundError as e:
        logger.error(e)
        return errno.ENOENT

    subprocrunner.SubprocessRunner.is_save_history = True
    if options.tc_command_output != TcCommandOutput.NOT_SET:
        subprocrunner.SubprocessRunner.default_is_dry_run = True

    tc_param = {}
    for device in options.device:
        try:
            verify_network_interface(device)
        except NetworkInterfaceNotFoundError as e:
            logger.debug(str(e))
            continue

        tc_param.update(TcShapingRuleParser(device, options.ip_version,
                                            logger).get_tc_parameter())

    command_history = "\n".join(subprocrunner.SubprocessRunner.get_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return 0

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        write_tc_script(
            TcCommand.TCSHOW,
            command_history,
            filename_suffix="-".join(options.device))
        return 0

    logger.debug("command history\n{}".format(command_history))
    print(json.dumps(tc_param, indent=4))

    return 0


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import unicode_literals

import random

import pytest

from tcconfig._iptables import (
    IptablesMangleController,
    IptablesMangleMark,
    VALID_CHAIN_LIST,
)


_DEF_SRC = "192.168.0.0/24"
_DEF_DST = "192.168.100.0/24"


prerouting_mangle_mark_list = [
    IptablesMangleMark(
        ip_version=4,
        line_number=1,
        mark_id=1,
        source=_DEF_SRC,
        destination=_DEF_DST,
        chain="PREROUTING",
        protocol="all"
    ),
]

input_mangle_mark_list = [
    IptablesMangleMark(
        ip_version=4,
        line_number=1,
        mark_id=1234,
        source="anywhere",
        destination=_DEF_DST,
        chain="INPUT",
        protocol="all"
    ),
]

output_mangle_mark_list = [
    IptablesMangleMark(
        ip_version=4,
        line_number=1,
        mark_id=12,
        source=_DEF_SRC,
        destination=_DEF_DST,
        chain="OUTPUT",
        protocol="tcp"
    ),
    IptablesMangleMark(
        ip_version=4,
        line_number=2,
        mark_id=123,
        source=_DEF_SRC,
        destination="anywhere",
        chain="OUTPUT",
        protocol="all"
    ),
    IptablesMangleMark(
        ip_version=4,
        line_number=3,
        mark_id=12345,
        source="anywhere",
        destination="anywhere",
        chain="OUTPUT",
        protocol="all"
    ),
]

mangle_mark_list = (
    prerouting_mangle_mark_list +
    input_mangle_mark_list +
    output_mangle_mark_list
)

reverse_mangle_mark_list = (
    list(reversed(prerouting_mangle_mark_list)) +
    list(reversed(input_mangle_mark_list)) +
    list(reversed(output_mangle_mark_list))
)


@pytest.fixture
def iptables_ctrl_ipv4():
    return IptablesMangleController(True, ip_version=4)


class Test_IptablesMangleMark_repr(object):

    def test_smoke(self):
        for mangle_mark in mangle_mark_list:
            assert len(str(mangle_mark)) > 0


class Test_IptablesMangleMark_to_append_command(object):
    _CMD_PREFIX = "iptables -A {:s} -t mangle -j MARK"

    @pytest.mark.parametrize(
        [
            "mark_id", "source", "destination", "chain", "protocol",
            "line_number", "expected"
        ],
        [
            [
                2, _DEF_SRC, _DEF_DST, "PREROUTING", "all", None,
                "{} --set-mark 2 -p all -s {} -d {}".format(
                    _CMD_PREFIX.format("PREROUTING"), _DEF_SRC, _DEF_DST),
            ],
            [
                2, _DEF_SRC, _DEF_DST, "OUTPUT", "all", 1,
                "{} --set-mark 2 -p all -s {} -d {}".format(
                    _CMD_PREFIX.format("OUTPUT"), _DEF_SRC, _DEF_DST),
            ],
            [
                2, _DEF_SRC, _DEF_DST, "OUTPUT", "tcp", 1,
                "{} --set-mark 2 -p tcp -s {} -d {}".format(
                    _CMD_PREFIX.format("OUTPUT"), _DEF_SRC, _DEF_DST),
            ],
            [
                100, _DEF_SRC, "anywhere", "INPUT", "all", 100,
                "{} --set-mark 100 -p all -s {}".format(
                    _CMD_PREFIX.format("INPUT"), _DEF_SRC),
            ],
            [
                1, "anywhere", _DEF_DST, "OUTPUT", "all", 100,
                "{} --set-mark 1 -p all -d {}".format(
                    _CMD_PREFIX.format("OUTPUT"), _DEF_DST),
            ],
            [
                1, "anywhere", "anywhere", "OUTPUT", "all", 100,
                "{} --set-mark 1 -p all".format(
                    _CMD_PREFIX.format("OUTPUT")),
            ],
        ]
    )
    def test_normal(
            self, mark_id, source, destination, chain, protocol, line_number,
            expected):
        mark = IptablesMangleMark(
            ip_version=4,
            mark_id=mark_id, source=source, destination=destination,
            chain=chain, protocol=protocol, line_number=line_number)
        assert mark.to_append_command() == expected


class Test_IptablesMangleMark_to_delete_command(object):

    @pytest.mark.parametrize(
        [
            "mark_id", "source", "destination",  "chain", "protocol",
            "line_number", "expected"
        ],
        [
            [
                2, _DEF_SRC, _DEF_DST, "PREROUTING", "all", 1,
                "iptables -t mangle -D PREROUTING 1",
            ],
            [
                20, None, None, "OUTPUT", "all", 2,
                "iptables -t mangle -D OUTPUT 2",
            ],
        ]
    )
    def test_normal(
            self, mark_id, source, destination, chain, protocol, line_number,
            expected):
        mark = IptablesMangleMark(
            ip_version=4,
            mark_id=mark_id, source=_DEF_SRC, destination=_DEF_DST,
            chain=chain, protocol=protocol, line_number=line_number)
        assert mark.to_delete_command() == expected

    @pytest.mark.parametrize(
        [
            "mark_id", "source", "destination", "chain", "protocol",
            "line_number", "expected"
        ],
        [
            [
                2, _DEF_SRC, _DEF_DST, "OUTPUT", "all", None,
                TypeError,
            ],
        ]
    )
    def test_exception(
            self, mark_id, source, destination, chain, protocol, line_number,
            expected):
        mark = IptablesMangleMark(
            ip_version=4,
            mark_id=mark_id, source=source, destination=destination,
            chain=chain, protocol=protocol, line_number=line_number)
        with pytest.raises(expected):
            mark.to_delete_command()


class Test_IptablesMangleController_get_unique_mark_id(object):

    @pytest.mark.xfail
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        for i in range(5):
            mark_id = iptables_ctrl_ipv4.get_unique_mark_id()

            assert mark_id == (i + 101)

            mangle_mark = IptablesMangleMark(
                ip_version=4,
                mark_id=mark_id, source=_DEF_SRC, destination=_DEF_DST,
                chain=random.choice(VALID_CHAIN_LIST))

            assert iptables_ctrl_ipv4.add(mangle_mark) == 0


class Test_IptablesMangleController_add(object):

    @pytest.mark.xfail
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()
        initial_len = len(iptables_ctrl_ipv4.get_iptables())

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        assert len(iptables_ctrl_ipv4.get_iptables()) > initial_len


class Test_IptablesMangleController_clear(object):

    @pytest.mark.xfail
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        initial_len = len(iptables_ctrl_ipv4.get_iptables())

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        assert len(iptables_ctrl_ipv4.get_iptables()) > initial_len

        iptables_ctrl_ipv4.clear()

        assert len(iptables_ctrl_ipv4.get_iptables()) == initial_len


class Test_IptablesMangleController_parse(object):
    @pytest.mark.xfail
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        for lhs_mangle, rhs_mangle in zip(
                iptables_ctrl_ipv4.parse(), reverse_mangle_mark_list):

            print("lhs: {}".format(lhs_mangle))
            print("rhs: {}".format(rhs_mangle))

            assert lhs_mangle == rhs_mangle



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

import errno
import ipaddress
import sys

import logbook
import six
import subprocrunner
import typepy

import pyparsing as pp

from ._argparse_wrapper import ArgparseWrapper
from ._common import (
    check_tc_command_installation,
    is_execute_tc_command,
    write_tc_script,
)
from ._const import (
    VERSION,
    Network,
    Tc,
    TcCommand,
    TcCommandOutput,
    TrafficDirection,
)
from ._error import (
    ModuleNotFoundError,
    NetworkInterfaceNotFoundError,
    UnitNotFoundError,
)
from ._logger import (
    LOG_FORMAT_STRING,
    logger,
    set_log_level,
)
from .parser.shaping_rule import TcShapingRuleParser
from .traffic_control import TrafficControl


logbook.StderrHandler(
    level=logbook.DEBUG, format_string=LOG_FORMAT_STRING).push_application()


def parse_option():
    parser = ArgparseWrapper(VERSION)

    group = parser.parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--device",
        help="network device name (e.g. eth0)")
    group.add_argument(
        "-f", "--config-file",
        help="""setting traffic controls from a configuration file.
        output file of the tcshow.""")

    group = parser.parser.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite", action="store_true", default=False,
        help="overwrite existing traffic shaping rules.")
    group.add_argument(
        "--change", dest="is_change_shaper", action="store_true",
        default=False,
        help="""
        change existing traffic shaping rules to new one. this option reduce
        the shaping rule switching side effect (such as traffic spike)
        compared to --overwrite option.
        note: the tcset command will fail when there is no existing shaping
        rules.
        """)
    group.add_argument(
        "--add", dest="is_add_shaper", action="store_true", default=False,
        help="add a traffic shaping rule in addition to existing rules.")

    group = parser.parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--direction", choices=TrafficDirection.LIST,
        default=TrafficDirection.OUTGOING,
        help="""the direction of network communication that impose traffic control.
        ``incoming`` requires Linux kernel version 2.6.20 or later.
        (default = ``%(default)s``)
        """)
    group.add_argument(
        "--rate", "--bandwidth-rate", dest="bandwidth_rate",
        help="""
        network bandwidth rate [bit per second].
        valid units are either: K/M/G/Kbps/Mbps/Gbps
        e.g. --rate 10Mbps
        """)
    group.add_argument(
        "--delay", dest="network_latency", type=float, default=0,
        help="""
        round trip network delay [ms]. the valid range is {:d} to {:d}.
        (default=%(default)s)
        """.format(
            TrafficControl.MIN_LATENCY_MS, TrafficControl.MAX_LATENCY_MS))
    group.add_argument(
        "--delay-distro", dest="latency_distro_ms", type=float, default=0,
        help="""
        distribution of network latency becomes X +- Y [ms]
        (normal distribution). Here X is the value of --delay option and
        Y is the value of --delay-dist option).
        network latency distribution will be uniform without this option.
        """)
    group.add_argument(
        "--loss", dest="packet_loss_rate", type=float, default=0,
        help="""
        round trip packet loss rate [%%]. the valid range is {:d} to {:d}.
        (default=%(default)s)
        """.format(
            TrafficControl.MIN_PACKET_LOSS_RATE,
            TrafficControl.MAX_PACKET_LOSS_RATE))
    group.add_argument(
        "--duplicate", dest="packet_duplicate_rate", type=float, default=0,
        help="""
        round trip packet duplicate rate [%%]. the valid range is {:d} to {:d}.
        (default=%(default)s)
        """.format(
            TrafficControl.MIN_PACKET_DUPLICATE_RATE,
            TrafficControl.MAX_PACKET_DUPLICATE_RATE))
    group.add_argument(
        "--corrupt", dest="corruption_rate", type=float, default=0,
        help="""
        packet corruption rate [%%]. the valid range is {:d} to {:d}.
        packet corruption means single bit error at a random offset in
        the packet. (default=%(default)s)
        """.format(
            TrafficControl.MIN_CORRUPTION_RATE,
            TrafficControl.MAX_CORRUPTION_RATE))
    group.add_argument(
        "--reordering", dest="reordering_rate", type=float, default=0,
        help="""
        packet reordering rate [%%]. the valid range is {:d} to {:d}.
        (default=%(default)s)
        """.format(
            TrafficControl.MIN_REORDERING_RATE,
            TrafficControl.MAX_REORDERING_RATE))
    group.add_argument(
        "--network", "--dst-network", dest="dst_network",
        help="target IP address/network to control traffic")
    group.add_argument(
        "--port", "--dst-port", dest="dst_port", type=int,
        help="target destination port number to control traffic.")
    group.add_argument(
        "--src-port", type=int,
        help="target source port number to control traffic.")
    group.add_argument(
        "--ipv6", dest="is_ipv6", action="store_true", default=False,
        help="apply traffic control to IPv6 packets rather than IPv4.")
    group.add_argument(
        "--shaping-algo", dest="shaping_algorithm",
        choices=["tbf", "htb"], default="htb",
        help="shaping algorithm. defaults to %(default)s (recommended).")

    group = parser.parser.add_argument_group("Routing")
    group.add_argument(
        "--iptables", dest="is_enable_iptables",
        action="store_true", default=False,
        help="use iptables to traffic shaping.")
    group.add_argument(
        "--src-network",
        help="""
        set traffic shaping rule to a specific packets that routed from
        --src-network to --dst-network. This option required to execute with
        the --iptables option.
        the shaping rule only affect to outgoing packets
        (no effect to if you execute with "--direction incoming" option)
        """)

    return parser.parser.parse_args()


def verify_netem_module():
    import re

    runner = subprocrunner.SubprocessRunner("lsmod")
    if runner.run() != 0:
        raise OSError("failed to execute lsmod")

    if re.search(r"\bsch_netem\b", runner.stdout) is None:
        raise ModuleNotFoundError("sch_netem module not found")


class TcConfigLoader(object):
    def __init__(self, logger):
        self.__logger = logger
        self.__config_table = None
        self.is_overwrite = False

    def load_tcconfig(self, config_file_path):
        import json
        from voluptuous import Schema, Required, Any, ALLOW_EXTRA

        schema = Schema({
            Required(six.text_type): {
                Any(*TrafficDirection.LIST): {
                    six.text_type: {
                        six.text_type: six.text_type,
                    },
                }
            },
        }, extra=ALLOW_EXTRA)

        with open(config_file_path) as fp:
            self.__config_table = json.load(fp)

        schema(self.__config_table)
        self.__logger.debug("tc config file: {:s}".format(
            json.dumps(self.__config_table, indent=4)))

    def get_tcconfig_command_list(self):
        command_list = []

        for device, device_table in six.iteritems(self.__config_table):
            device_option = "--device={:s}".format(device)

            if self.is_overwrite:
                command_list.append("{:s} {:s}".format(
                    TcCommand.TCDEL, device_option))

            for direction, direction_table in six.iteritems(device_table):
                is_first_set = True

                for tc_filter, filter_table in six.iteritems(direction_table):
                    self.__logger.debug(
                        "is_first_set={}, filter='{}', table={}".format(
                            is_first_set, tc_filter, filter_table))

                    if not filter_table:
                        continue

                    option_list = [
                        device_option,
                        "--direction={:s}".format(direction),
                    ] + [
                        "--{:s}={:s}".format(k, v)
                        for k, v in six.iteritems(filter_table)
                    ]

                    try:
                        dst_network = self.__parse_tc_filter_network(tc_filter)
                        if dst_network not in (
                                Network.Ipv4.ANYWHERE, Network.Ipv6.ANYWHERE):
                            option_list.append(
                                "--dst-network={:s}".format(dst_network))
                    except pp.ParseException:
                        pass

                    try:
                        src_port = self.__parse_tc_filter_src_port(tc_filter)
                        option_list.append("--src-port={}".format(src_port))
                    except pp.ParseException:
                        pass

                    try:
                        dst_port = self.__parse_tc_filter_dst_port(tc_filter)
                        option_list.append("--dst-port={}".format(dst_port))
                    except pp.ParseException:
                        pass

                    if not is_first_set:
                        option_list.append("--add")

                    is_first_set = False

                    command_list.append(
                        " ".join([TcCommand.TCSET] + option_list))

        return command_list

    @staticmethod
    def __parse_tc_filter_network(text):
        network_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.DST_NETWORK), include=True) +
            pp.Word(pp.alphanums + "." + "/"))

        return network_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_src_port(text):
        port_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.SRC_PORT), include=True) +
            pp.Word(pp.nums))

        return port_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_dst_port(text):
        port_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.DST_PORT), include=True) +
            pp.Word(pp.nums))

        return port_pattern.parseString(text)[-1]


class TcShapingRuleFinder(object):
    @property
    def tc(self):
        return self.__tc

    def __init__(self, device, tc):
        self.__device = device
        self.__tc = tc

    def exist_rule(self):
        parser = TcShapingRuleParser(
            self.__device, self.tc.ip_version, logger)

        key_param_list = (
            Tc.Param.DST_NETWORK, Tc.Param.SRC_NETWORK,
            Tc.Param.DST_PORT, Tc.Param.SRC_NETWORK,
        )

        new_tc_filter = {
            Tc.Param.DST_NETWORK: self.tc.dst_network,
            Tc.Param.SRC_NETWORK: self.tc.src_network,
            Tc.Param.DST_PORT: self.tc.dst_port,
            Tc.Param.SRC_NETWORK: self.tc.src_port,
        }

        if self.tc.direction == TrafficDirection.OUTGOING:
            current_tc_filter_list = parser.get_outgoing_tc_filter()
        elif self.tc.direction == TrafficDirection.INCOMING:
            current_tc_filter_list = parser.get_incoming_tc_filter()

        logger.debug(
            "exist_rule: direction={}, new-filter={} current-filters={}".format(
                self.tc.direction, new_tc_filter, current_tc_filter_list))

        for cuurent_tc_filter in current_tc_filter_list:
            if all([
                cuurent_tc_filter.get(
                    key_param) == new_tc_filter.get(key_param)
                for key_param in key_param_list
            ]):
                logger.debug("existing shaping rule found: {}".format(
                    cuurent_tc_filter))
                return True

        return False


def set_tc_from_file(logger, config_file_path, is_overwrite):
    return_code = 0

    loader = TcConfigLoader(logger)
    loader.is_overwrite = is_overwrite
    loader.load_tcconfig(config_file_path)

    for tcconfig_command in loader.get_tcconfig_command_list():
        return_code |= subprocrunner.SubprocessRunner(
            tcconfig_command).run()

    return return_code


def main():
    options = parse_option()

    set_log_level(options.log_level)

    if is_execute_tc_command(options.tc_command_output):
        check_tc_command_installation()
    else:
        subprocrunner.SubprocessRunner.default_is_dry_run = True

    try:
        verify_netem_module()
    except ModuleNotFoundError as e:
        logger.debug(str(e))
    except subprocrunner.CommandNotFoundError as e:
        logger.error(str(e))

    if typepy.is_not_null_string(options.config_file):
        return set_tc_from_file(logger, options.config_file, options.overwrite)

    subprocrunner.SubprocessRunner.is_save_history = True

    try:
        tc = TrafficControl(
            options.device,
            direction=options.direction,
            bandwidth_rate=options.bandwidth_rate,
            latency_ms=options.network_latency,
            latency_distro_ms=options.latency_distro_ms,
            packet_loss_rate=options.packet_loss_rate,
            packet_duplicate_rate=options.packet_duplicate_rate,
            corruption_rate=options.corruption_rate,
            reordering_rate=options.reordering_rate,
            dst_network=options.dst_network,
            src_network=options.src_network,
            src_port=options.src_port,
            dst_port=options.dst_port,
            is_ipv6=options.is_ipv6,
            is_change_shaper=options.is_change_shaper,
            is_add_shaper=options.is_add_shaper,
            is_enable_iptables=options.is_enable_iptables,
            shaping_algorithm=options.shaping_algorithm,
            tc_command_output=options.tc_command_output,
        )
    except UnitNotFoundError:
        logger.error(
            "--rate/--bandwidth-rate require unit such as K/M/Kbps/Mbps/etc.")
        return errno.EINVAL

    try:
        tc.validate()
    except (NetworkInterfaceNotFoundError) as e:
        logger.error(str(e))
        return errno.EINVAL
    except ipaddress.AddressValueError as e:
        logger.error(
            "{}. ".format(e) +
            "--ipv6 option will be required to use IPv6 address.")
        return errno.EINVAL
    except ValueError as e:
        logger.error(e)
        return errno.EINVAL

    try:
        tc.sanitize()
    except ValueError as e:
        logger.error(e)
        return errno.EINVAL

    if options.overwrite:
        if options.log_level == logbook.INFO:
            set_log_level(logbook.ERROR)

        try:
            tc.delete_tc()
        except NetworkInterfaceNotFoundError:
            pass

        set_log_level(options.log_level)

    if (
        options.is_add_shaper and
        TcShapingRuleFinder(device=options.device, tc=tc).exist_rule()
    ):
        logger.error(
            "adding a shaping rule failed. a shaping rule for the same "
            "network/port already exist. try --overwrite option if you want "
            "to overwrite the existing rule.")
        return errno.EINVAL

    return_code = tc.set_tc()
    command_history = "\n".join(tc.get_command_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return 0

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        write_tc_script(
            TcCommand.TCSET, command_history, filename_suffix=options.device)
        return 0

    logger.debug("command history\n{}".format(command_history))

    return return_code


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import ipaddress
import re

import typepy

import pyparsing as pp

from .._const import Tc
from .._logger import logger
from ._common import _to_unicode


class TcFilterParser(object):

    class FilterMatchIdIpv4(object):
        INCOMING_NETWORK = 12
        OUTGOING_NETWORK = 16
        PORT = 20

    class FilterMatchIdIpv6(object):
        INCOMING_NETWORK_LIST = [8, 12, 16, 20]
        OUTGOING_NETWORK_LIST = [24, 28, 32, 36]
        PORT = 40

    __FILTER_FLOWID_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("flowid", include=True) +
        pp.Word(pp.hexnums + ":")
    )
    __FILTER_PROTOCOL_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("protocol", include=True) +
        pp.Word(pp.alphanums)
    )
    __FILTER_MATCH_PATTERN = (
        pp.Literal("match") +
        pp.Word(pp.alphanums + "/") +
        pp.Literal("at") +
        pp.Word(pp.nums)
    )
    __FILTER_MANGLE_MARK_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("handle", include=True) +
        pp.Word(pp.hexnums) +
        pp.SkipTo("classid", include=True) +
        pp.Word(pp.hexnums + ":")
    )

    @property
    def protocol(self):
        return self.__protocol

    def __init__(self, ip_version):
        self.__ip_version = ip_version
        self.__clear()

        self.__buffer = None
        self.__parse_idx = 0

        self.__protocol = None

    def parse_filter(self, text):
        self.__clear()

        if typepy.is_null_string(text):
            return []

        filter_data_matrix = []
        self.__buffer = text.splitlines()
        self.__parse_idx = 0

        while self.__parse_idx < len(self.__buffer):
            line = self.__buffer[self.__parse_idx].strip()
            self.__parse_idx += 1

            if typepy.is_null_string(line):
                continue

            try:
                self.__parse_mangle_mark(line)
            except pp.ParseException:
                logger.debug("failed to parse mangle: {}".format(line))
            else:
                filter_data_matrix.append({
                    Tc.Param.CLASS_ID: self.__classid,
                    Tc.Param.HANDLE: self.__handle,
                })
                self.__clear()
                continue

            tc_filter = self.__get_filter()

            try:
                self.__parse_flow_id(line)

                if tc_filter.get(Tc.Param.FLOW_ID):
                    logger.debug("store filter: {}".format(tc_filter))
                    filter_data_matrix.append(tc_filter)
                    self.__clear()
                    self.__parse_flow_id(line)

                continue
            except pp.ParseException:
                logger.debug("failed to parse flow id: {}".format(line))

            try:
                self.__parse_protocol(line)
                continue
            except pp.ParseException:
                logger.debug("failed to parse protocol: {}".format(line))

            try:
                if self.__ip_version == 4:
                    self.__parse_filter_ipv4(line)
                elif self.__ip_version == 6:
                    self.__parse_filter_ipv6(line)
                else:
                    raise ValueError(
                        "unknown ip version: {}".format(self.__ip_version))
            except pp.ParseException:
                logger.debug("failed to parse filter: {}".format(line))

        if self.__flow_id:
            filter_data_matrix.append(self.__get_filter())

        return filter_data_matrix

    def parse_incoming_device(self, text):
        if typepy.is_null_string(text):
            return None

        match = re.search(
            "Egress Redirect to device ifb[\d]+",
            _to_unicode(text), re.MULTILINE)
        if match is None:
            return None

        return re.search("ifb[\d]+", match.group()).group()

    def __clear(self):
        self.__flow_id = None
        self.__filter_network = None
        self.__filter_src_port = None
        self.__filter_dst_port = None

        self.__handle = None
        self.__classid = None

    def __get_filter(self):
        return {
            Tc.Param.FLOW_ID: self.__flow_id,
            Tc.Param.DST_NETWORK: self.__filter_network,
            Tc.Param.SRC_PORT: self.__filter_src_port,
            Tc.Param.DST_PORT: self.__filter_dst_port,
            Tc.Param.PROTOCOL: self.protocol
        }

    def __parse_flow_id(self, line):
        parsed_list = self.__FILTER_FLOWID_PATTERN.parseString(
            _to_unicode(line.lstrip()))
        self.__flow_id = parsed_list[-1]
        logger.debug("succeed to parse flow id: flow-id={}, line={}".format(
            self.__flow_id, line))

    def __parse_protocol(self, line):
        parsed_list = self.__FILTER_PROTOCOL_PATTERN.parseString(
            _to_unicode(line.lstrip()))
        self.__protocol = parsed_list[-1]

    def __parse_mangle_mark(self, line):
        parsed_list = self.__FILTER_MANGLE_MARK_PATTERN.parseString(
            _to_unicode(line.lstrip()))
        self.__classid = parsed_list[-1]
        self.__handle = int(u"0" + parsed_list[-3], 16)
        logger.debug(
            "succeed to parse mangle mark: "
            "classid={}, handle={}, line={}".format(
                self.__classid, self.__handle, line))

    def __parse_filter_line(self, line):
        parsed_list = self.__FILTER_MATCH_PATTERN.parseString(
            _to_unicode(line))
        value_hex, mask_hex = parsed_list[1].split("/")
        match_id = int(parsed_list[3])

        return (value_hex, mask_hex, match_id)

    def __parse_filter_port(self, value_hex):
        # Port filter consists eight hex digits.
        # The upper-half represents source port filter and
        # the bottom-half represents destination port filter.

        if len(value_hex) != 8:
            raise ValueError("invalid port filter value: {}".format(value_hex))

        src_port_hex = value_hex[:4]
        dst_port_hex = value_hex[4:]

        logger.debug(
            "parse ipv4 port: src-port-hex={}, dst-port-hex={}".format(
                src_port_hex, dst_port_hex))

        src_port_decimal = int(src_port_hex, 16)
        self.__filter_src_port = (
            src_port_decimal if src_port_decimal != 0 else None)

        dst_port_decimal = int(dst_port_hex, 16)
        self.__filter_dst_port = (
            dst_port_decimal if dst_port_decimal != 0 else None)

    def __parse_filter_ipv4(self, line):
        value_hex, mask_hex, match_id = self.__parse_filter_line(line)

        if match_id in [
            self.FilterMatchIdIpv4.INCOMING_NETWORK,
            self.FilterMatchIdIpv4.OUTGOING_NETWORK,
        ]:
            ipaddr = ".".join([
                str(int(value_hex[i: i + 2], 16))
                for i in range(0, len(value_hex), 2)
            ])
            netmask = bin(int(mask_hex, 16)).count("1")

            self.__filter_network = "{:s}/{:d}".format(ipaddr, netmask)
        elif match_id == self.FilterMatchIdIpv4.PORT:
            self.__parse_filter_port(value_hex)

        logger.debug(
            "succeed to parse filter: " + ", ".join([
                "dst_network={}".format(self.__filter_network),
                "src_port={}".format(self.__filter_src_port),
                "dst_port={}".format(self.__filter_dst_port),
                "line={}".format(line)
            ]))

    def __parse_filter_ipv6(self, line):
        netmask = 0
        value_hex, mask_hex, match_id = self.__parse_filter_line(line)

        octet_list = []
        octet_len = 4

        if (
            match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST or
            match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST
        ):
            octet_list.extend([
                value_hex[i: i + octet_len]
                for i in range(0, len(value_hex), octet_len)
            ])
            netmask += bin(int(mask_hex, 16)).count("1")

            while True:
                try:
                    line = self.__buffer[self.__parse_idx].strip()
                except IndexError:
                    break

                try:
                    value_hex, mask_hex, match_id = self.__parse_filter_line(
                        line)
                except pp.ParseException:
                    break

                if (
                    match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST or
                    match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST
                ):
                    octet_list.extend([
                        value_hex[i: i + octet_len]
                        for i in range(0, len(value_hex), octet_len)
                    ])
                    netmask += bin(int(mask_hex, 16)).count("1")
                else:
                    break

                self.__parse_idx += 1

            while len(octet_list) < 8:
                octet_list.append("0000")

            self.__filter_network = ipaddress.IPv6Network("{:s}/{:d}".format(
                ":".join(octet_list), netmask)).compressed

        elif match_id == self.FilterMatchIdIpv6.PORT:
            self.__parse_filter_port(value_hex)

        logger.debug(
            "succeed to parse filter: " + ", ".join([
                "dst_network={}".format(self.__filter_network),
                "src_port={}".format(self.__filter_src_port),
                "dst_port={}".format(self.__filter_dst_port),
                "line={}".format(line)
            ]))



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals


VERSION = "0.12.0"

KILO_SIZE = 1000
LIST_MANGLE_TABLE_COMMAND = "iptables -t mangle --line-numbers -L"


class Network(object):
    class Ipv4(object):
        ANYWHERE = "0.0.0.0/0"

    class Ipv6(object):
        ANYWHERE = "::0/0"


class Tc(object):

    class Subcommand(object):
        CLASS = "class"
        FILTER = "filter"
        QDISC = "qdisc"
        SHOW = "show"

    class Param(object):
        CLASS_ID = "classid"
        DST_NETWORK = "dst-network"
        DST_PORT = "dst-port"
        FLOW_ID = "flowid"
        HANDLE = "handle"
        PARENT = "parent"
        PROTOCOL = "protocol"
        SRC_NETWORK = "src-network"
        SRC_PORT = "src-port"


class TcCommand(object):
    TCSET = "tcset"
    TCDEL = "tcdel"
    TCSHOW = "tcshow"


class TcCommandOutput(object):
    NOT_SET = None
    STDOUT = "STDOUT"
    SCRIPT = "SCRIPT"


class TrafficDirection(object):
    OUTGOING = "outgoing"
    INCOMING = "incoming"
    LIST = [OUTGOING, INCOMING]



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

import errno
import sys

import logbook
import subprocrunner

from ._argparse_wrapper import ArgparseWrapper
from ._common import (
    check_tc_command_installation,
    is_execute_tc_command,
    verify_network_interface,
    write_tc_script,
)
from ._const import (
    VERSION,
    TcCommand,
    TcCommandOutput,
)
from ._error import NetworkInterfaceNotFoundError
from ._logger import (
    LOG_FORMAT_STRING,
    logger,
    set_logger,
    set_log_level,
)
from .traffic_control import TrafficControl


logbook.StderrHandler(
    level=logbook.DEBUG, format_string=LOG_FORMAT_STRING).push_application()


def parse_option():
    parser = ArgparseWrapper(VERSION)

    group = parser.parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--device", required=True,
        help="network device name (e.g. eth0)")

    return parser.parser.parse_args()


def main():
    options = parse_option()

    set_log_level(options.log_level)

    if is_execute_tc_command(options.tc_command_output):
        check_tc_command_installation()
    else:
        subprocrunner.SubprocessRunner.default_is_dry_run = True
        set_logger(False)

    try:
        verify_network_interface(options.device)
    except NetworkInterfaceNotFoundError as e:
        logger.error(e)
        return errno.EINVAL

    subprocrunner.SubprocessRunner.is_save_history = True

    tc = TrafficControl(options.device)
    if options.log_level == logbook.INFO:
        subprocrunner.set_log_level(logbook.ERROR)

    try:
        return_code = tc.delete_tc()
    except NetworkInterfaceNotFoundError as e:
        logger.debug(e)
        return 0

    command_history = "\n".join(tc.get_command_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return return_code

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        set_logger(True)
        write_tc_script(
            TcCommand.TCDEL, command_history, filename_suffix=options.device)
        return return_code

    logger.debug("command history\n{}".format(command_history))

    return return_code


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import division
from __future__ import unicode_literals

import pytest
from subprocrunner import SubprocessRunner
import typepy

from allpairspy import AllPairs
from tcconfig._const import TcCommand


SKIP_TEST = False


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


def is_valid_combination(row):
    if all([typepy.is_null_string(param) for param in row]):
        return False

    return True


def is_invalid_param(rate, delay, loss, corrupt):
    params = [
        rate,
        delay,
        loss,
        corrupt,
    ]

    return all([
        typepy.is_null_string(param) for param in params
    ])


class NormalTestValue(object):
    RATE_LIST = [
        "",
        "--rate 100K",
        "--rate 0.5M",
    ]
    DELAY_LIST = [
        "",
        "--delay 100",
    ]
    DELAY_DISTRO_LIST = [
        "",
        "--delay-distro 20",
    ]
    PACKET_LOSS_RATE_LIST = [
        "",
        "--loss 0.1",
    ]
    CORRUPTION_RATE_LIST = [
        "",
        "--corrupt 0.1",
    ]
    DIRECTION_LIST = [
        "",
        "--direction outgoing",
        "--direction incoming",
    ]
    NETWORK_LIST = [
        "",
        "--network 192.168.0.10",
        "--network 192.168.0.0/24",
    ]
    PORT_LIST = [
        "",
        "--port 80",
    ]
    OVERWRITE_LIST = [
        "",
        "--add",
        "--overwrite",
    ]
    IPTABLES_LIST = [
        "",
        "--iptables",
    ]


class Test_tcconfig(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:
      python setup.py test --addopts "--runxfail --device=<test device>"

    These tests are expected to execute on following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.skipif("SKIP_TEST is True")
    @pytest.mark.parametrize(
        [
            "rate", "delay", "delay_distro", "loss", "corrupt",
            "direction", "network", "port", "overwrite", "is_enable_iptables",
        ],
        [
            opt_list
            for opt_list in AllPairs([
                NormalTestValue.RATE_LIST,
                NormalTestValue.DELAY_LIST,
                NormalTestValue.DELAY_DISTRO_LIST,
                NormalTestValue.PACKET_LOSS_RATE_LIST,
                NormalTestValue.CORRUPTION_RATE_LIST,
                NormalTestValue.DIRECTION_LIST,
                NormalTestValue.NETWORK_LIST,
                NormalTestValue.PORT_LIST,
                NormalTestValue.OVERWRITE_LIST,
                NormalTestValue.IPTABLES_LIST,
            ], n=3, filter_func=is_valid_combination)
        ])
    def test_smoke(
            self, device_value, rate, delay, delay_distro, loss, corrupt,
            direction, network, port, overwrite, is_enable_iptables):

        if device_value is None:
            pytest.skip("device is empty")

        if is_invalid_param(rate, delay, loss, corrupt):
            pytest.skip("skip null parameters")

        device_option = "--device {}".format(device_value)

        SubprocessRunner("{:s} {:s}".format(
            TcCommand.TCDEL, device_option)).run()

        tcset_proc = SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            rate, delay, delay_distro, loss, corrupt,
            direction, network, port, overwrite, is_enable_iptables,
        ]))
        assert tcset_proc.run() == 0, tcset_proc.stderr

        SubprocessRunner("tcdel {:s}".format(device_option)).run()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import copy

import subprocrunner
import typepy
from typepy.type import Integer

from .._common import run_tc_show
from .._const import (
    Tc,
    TrafficDirection,
)
from .._iptables import IptablesMangleController
from .._logger import logger
from ._class import TcClassParser
from ._filter import TcFilterParser
from ._qdisc import TcQdiscParser


class TcShapingRuleParser(object):
    @property
    def device(self):
        return self.__device

    def __init__(self, device, ip_version, logger):
        self.__device = device
        self.__ip_version = ip_version
        self.__logger = logger

        self.__iptables_ctrl = IptablesMangleController(True, ip_version)

    def get_tc_parameter(self):
        return {
            self.device: {
                TrafficDirection.OUTGOING: self.__get_shaping_rule(
                    self.device),
                TrafficDirection.INCOMING: self.__get_shaping_rule(
                    self.__get_ifb_from_device()),
            },
        }

    def get_outgoing_tc_filter(self):
        return self.__parse_tc_filter(self.device)

    def get_incoming_tc_filter(self):
        ifb_device = self.__get_ifb_from_device()
        if ifb_device is None:
            return []

        return self.__parse_tc_filter(ifb_device)

    def __get_ifb_from_device(self):
        filter_runner = subprocrunner.SubprocessRunner(
            "tc filter show dev {:s} root".format(self.device), dry_run=False)
        filter_runner.run()

        return TcFilterParser(self.__ip_version).parse_incoming_device(
            filter_runner.stdout)

    def __get_filter_key(self, filter_param):
        network_format = Tc.Param.DST_NETWORK + "={:s}"
        protocol_format = Tc.Param.PROTOCOL + "={:s}"
        key_item_list = []

        if Tc.Param.HANDLE in filter_param:
            handle = filter_param.get(Tc.Param.HANDLE)
            Integer(handle).validate()
            handle = int(handle)

            for mangle in self.__iptables_ctrl.parse():
                if mangle.mark_id != handle:
                    continue

                key_item_list.append(network_format.format(mangle.destination))
                if typepy.is_not_null_string(mangle.source):
                    key_item_list.append("{:s}={:s}".format(
                        Tc.Param.SRC_NETWORK, mangle.source))
                key_item_list.append(protocol_format.format(mangle.protocol))

                break
            else:
                raise ValueError("mangle mark not found: {}".format(mangle))
        else:
            network = filter_param.get(Tc.Param.DST_NETWORK)
            if typepy.is_not_null_string(network):
                key_item_list.append(network_format.format(network))

            src_port = filter_param.get(Tc.Param.SRC_PORT)
            if Integer(src_port).is_type():
                port_format = Tc.Param.SRC_PORT + "={:d}"
                key_item_list.append(port_format.format(src_port))

            dst_port = filter_param.get(Tc.Param.DST_PORT)
            if Integer(dst_port).is_type():
                port_format = Tc.Param.DST_PORT + "={:d}"
                key_item_list.append(port_format.format(dst_port))

            protocol = filter_param.get(Tc.Param.PROTOCOL)
            if typepy.is_not_null_string(protocol):
                key_item_list.append(protocol_format.format(protocol))

        return ", ".join(key_item_list)

    def __get_shaping_rule(self, device):
        if typepy.is_null_string(device):
            return {}

        class_param_list = self.__parse_tc_class(device)
        filter_param_list = self.__parse_tc_filter(device)
        qdisc_param_list = self.__parse_tc_qdisc(device)

        shaping_rule_mapping = {}

        for filter_param in filter_param_list:
            logger.debug(
                "{:s} param: {}".format(Tc.Subcommand.FILTER, filter_param))
            shaping_rule = {}

            filter_key = self.__get_filter_key(filter_param)
            if typepy.is_null_string(filter_key):
                logger.debug("empty filter key: {}".format(filter_param))
                continue

            for qdisc_param in qdisc_param_list:
                logger.debug(
                    "{:s} param: {}".format(Tc.Subcommand.QDISC, qdisc_param))

                if qdisc_param.get(Tc.Param.PARENT) not in (
                        filter_param.get(Tc.Param.FLOW_ID),
                        filter_param.get(Tc.Param.CLASS_ID)):
                    continue

                work_qdisc_param = copy.deepcopy(qdisc_param)
                del work_qdisc_param[Tc.Param.PARENT]
                shaping_rule.update(work_qdisc_param)

            for class_param in class_param_list:
                logger.debug(
                    "{:s} param: {}".format(Tc.Subcommand.CLASS, class_param))

                if class_param.get(Tc.Param.CLASS_ID) not in (
                        filter_param.get(Tc.Param.FLOW_ID),
                        filter_param.get(Tc.Param.CLASS_ID)):
                    continue

                work_class_param = copy.deepcopy(class_param)
                del work_class_param[Tc.Param.CLASS_ID]
                shaping_rule.update(work_class_param)

            if not shaping_rule:
                continue

            logger.debug(
                "rule found: {} {}".format(filter_key, shaping_rule))

            shaping_rule_mapping[filter_key] = shaping_rule

        return shaping_rule_mapping

    def __parse_tc_qdisc(self, device):
        try:
            param_list = list(TcQdiscParser().parse(
                run_tc_show(Tc.Subcommand.QDISC, device)))
        except ValueError:
            return []

        logger.debug("tc qdisc parse result: {}".format(param_list))

        return param_list

    def __parse_tc_filter(self, device):
        param_list = list(TcFilterParser(self.__ip_version).parse_filter(
            run_tc_show(Tc.Subcommand.FILTER, device)))
        logger.debug("tc filter parse result: {}".format(param_list))

        return param_list

    def __parse_tc_class(self, device):
        param_list = list(TcClassParser().parse(
            run_tc_show(Tc.Subcommand.CLASS, device)))
        logger.debug("tc class parse result: {}".format(param_list))

        return param_list



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import re

import typepy

from ._common import _to_unicode


class TcClassParser(object):

    class Pattern(object):
        CLASS_ID = "[0-9a-z:]+"
        RATE = "[0-9]+[KMGT]?"

    class Key(object):
        CLASS_ID = "classid"
        RATE = "rate"

    def __init__(self):
        self.__clear()

    def parse(self, text):
        for line in text.splitlines():
            self.__clear()

            if typepy.is_null_string(line):
                continue

            line = _to_unicode(line.lstrip())

            self.__parse_classid(line)
            self.__parse_rate(line)

            yield self.__parsed_param

    def __clear(self):
        self.__parsed_param = {}

    def __parse_classid(self, line):
        self.__parsed_param[self.Key.CLASS_ID] = None
        tag = "class htb "

        match = re.search("{:s}{:s}".format(tag, self.Pattern.CLASS_ID), line)
        if match is None:
            return

        self.__parsed_param[self.Key.CLASS_ID] = re.search(
            self.Pattern.CLASS_ID, match.group().lstrip(tag)).group()

    def __parse_rate(self, line):
        self.__parsed_param[self.Key.RATE] = None

        match = re.search("rate {:s}".format(self.Pattern.RATE), line)
        if match is None:
            return

        self.__parsed_param[self.Key.RATE] = re.search(
            self.Pattern.RATE, match.group()).group()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division

import platform

import pingparsing
import pytest
from subprocrunner import SubprocessRunner
import typepy

from tcconfig._const import TcCommand

from .common import execute_tcdel


WAIT_TIME = 3  # [sec]


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-f -q"
    transmitter.waittime = WAIT_TIME

    return transmitter


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


class Test_tcset_one_network(object):
    """
    Tests in this class are not executable on Travis .
    Execute following command at the local environment  when running tests:
      python setup.py test --addopts \
          "--device=<test device> --dst-host=<hostname/IP-addr>"

    These tests are expected to execute on following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(["delay"], [
        [100],
    ])
    def test_const_latency(
            self, device_option, dst_host_option, transmitter, pingparser,
            delay):
        if device_option is None:
            pytest.skip("device option is null")

        if typepy.is_null_string(dst_host_option):
            pytest.skip("destination host is null")

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/ latency tc ---
        command_list = [
            TcCommand.TCSET,
            "--device " + device_option,
            "--delay {:d}".format(delay),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        result = transmitter.ping()
        pingparser.parse(result.stdout)
        with_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        # finalize ---
        execute_tcdel(device_option)

    @pytest.mark.skipif("platform.system() == 'Windows'")
    @pytest.mark.parametrize(["delay", "delay_distro"], [
        [100, 50],
    ])
    def test_const_latency_distro(
            self, device_option, dst_host_option, transmitter, pingparser,
            delay, delay_distro):
        if typepy.is_null_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        without_tc_rtt_avg = pingparser.rtt_avg
        without_tc_rtt_mdev = pingparser.rtt_mdev

        # w/ latency tc ---
        command_list = [
            TcCommand.TCSET,
            "--device " + device_option,
            "--delay {:d}".format(delay),
            "--delay-distro {:d}".format(delay_distro),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        result = transmitter.ping()
        pingparser.parse(result.stdout)
        with_tc_rtt_avg = pingparser.rtt_avg
        with_tc_rtt_mdev = pingparser.rtt_mdev

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        rtt_diff = with_tc_rtt_mdev - without_tc_rtt_mdev
        assert rtt_diff > (delay_distro / 2.0)

        # finalize ---
        execute_tcdel(device_option)

    @pytest.mark.parametrize(["option", "value"], [
        ["--loss", 10],
        ["--corrupt", 10],
    ])
    def test_const_packet_loss(
            self, device_option, dst_host_option, transmitter, pingparser,
            option, value):
        if typepy.is_null_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o packet loss tc ---
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        without_tc_loss_rate = pingparser.packet_loss_rate

        # w/ packet loss tc ---
        command_list = [
            TcCommand.TCSET,
            "--device " + device_option,
            "{:s} {:f}".format(option, value),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        result = transmitter.ping()
        pingparser.parse(result.stdout)
        with_tc_loss_rate = pingparser.packet_loss_rate

        # assertion ---
        loss_diff = with_tc_loss_rate - without_tc_loss_rate
        assert loss_diff > (value / 2)

        # finalize ---
        execute_tcdel(device_option)

    @pytest.mark.parametrize(["option", "value"], [
        ["--duplicate", 50],
    ])
    def test_const_packet_duplicate(
            self, device_option, dst_host_option, transmitter, pingparser,
            option, value):
        if typepy.is_null_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o packet duplicate tc ---
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        without_tc_duplicate_rate = pingparser.packet_duplicate_rate

        # w/ packet duplicate tc ---
        command_list = [
            TcCommand.TCSET,
            "--device {:s}".format(device_option),
            "{:s} {:f}".format(option, value),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        result = transmitter.ping()
        pingparser.parse(result.stdout)
        with_tc_duplicate_rate = pingparser.packet_duplicate_rate

        # assertion ---
        duplicate_rate_diff = with_tc_duplicate_rate - without_tc_duplicate_rate
        assert duplicate_rate_diff > (value / 2)

        # finalize ---
        execute_tcdel(device_option)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

import json

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import TcCommand


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcshow(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:
      python setup.py test --addopts "--runxfail --device=<test device>"
    """

    @pytest.mark.xfail
    def test_normal_ipv4(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        device_option = "--device {:s}".format(device_value)
        tcdel_command = "{:s} {:s}".format(TcCommand.TCDEL, device_option)

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--loss", "0.01",
            "--duplicate", "0.5",
            "--reorder", "0.2",
            "--rate", "0.25K",
            "--network", "192.168.0.10",
            "--port", "8080",
            "--overwrite",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "1",
            "--loss", "1",
            "--rate", "100M",
            "--network", "192.168.1.0/24",
            "--add",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--rate", "500K",
            "--direction", "incoming",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "1",
            "--loss", "0.02",
            "--duplicate", "0.5",
            "--reorder", "0.2",
            "--rate", "0.1M",
            "--network", "192.168.11.0/24",
            "--port", "80",
            "--direction", "incoming",
            "--add",
        ])).run() == 0

        runner = SubprocessRunner("{:s} {:s}".format(
            TcCommand.TCSHOW, device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
           "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                "delay": "10.0",
                "loss": "0.01",
                "duplicate": "0.5",
                "reorder": "0.2",
                "rate": "248",
                "delay-distro": "2.0"
            },
            "dst-network=192.168.1.0/24, protocol=ip": {
                "delay": "1.0",
                "loss": "1",
                "rate": "100M"
            }
        },
        "incoming": {
            "dst-network=192.168.11.0/24, dst-port=80, protocol=ip": {
                "delay": "1.0",
                "loss": "0.02",
                "duplicate": "0.5",
                "reorder": "0.2",
                "rate": "100K"
            },
            "dst-network=0.0.0.0/0, protocol=ip": {
                "delay": "10.0",
                "delay-distro": "2.0",
                "rate": "500K"
            }
        }
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))

        assert json.loads(runner.stdout) == json.loads(expected)

        SubprocessRunner(tcdel_command).run()

    @pytest.mark.xfail
    def test_normal_ipv6(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        device_option = "--device {:s}".format(device_value)
        tcdel_command = "{:s} {:s}".format(TcCommand.TCDEL, device_option)

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--loss", "0.01",
            "--duplicate", "5",
            "--reorder", "2",
            "--rate", "0.25K",
            "--network", "::1",
            "--port", "8080",
            "--overwrite",
            "--ipv6",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "1",
            "--loss", "1",
            "--rate", "100M",
            "--network", "2001:db00::0/24",
            "--add",
            "--ipv6",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--rate", "500K",
            "--direction", "incoming",
            "--ipv6",
        ])).run() == 0

        assert SubprocessRunner(" ".join([
            TcCommand.TCSET,
            device_option,
            "--delay", "1",
            "--loss", "0.02",
            "--duplicate", "5",
            "--reorder", "2",
            "--rate", "0.1M",
            "--network", "2001:db00::0/25",
            "--port", "80",
            "--direction", "incoming",
            "--add",
            "--ipv6",
        ])).run() == 0

        runner = SubprocessRunner("tcshow {:s} --ipv6".format(device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
            "dst-network=::1/128, dst-port=8080, protocol=ipv6": {
                "delay": "10.0",
                "loss": "0.01",
                "duplicate": "5",
                "reorder": "2",
                "rate": "248",
                "delay-distro": "2.0"
            },
            "dst-network=2001:db00::/24, protocol=ipv6": {
                "delay": "1.0",
                "loss": "1",
                "rate": "100M"
            }
        },
        "incoming": {
            "dst-network=2001:db00::/25, dst-port=80, protocol=ipv6": {
                "delay": "1.0",
                "loss": "0.02",
                "duplicate": "5",
                "reorder": "2",
                "rate": "100K"
            },
            "protocol=ipv6": {
                "delay": "10.0",
                "rate": "500K",
                "delay-distro": "2.0"
            }
        }
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))

        assert json.loads(runner.stdout) == json.loads(expected)

        SubprocessRunner(tcdel_command).run()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

import pytest

from tcconfig._common import (
    is_anywhere_network,
    get_anywhere_network,
    sanitize_network,
)


class Test_is_anywhere_network(object):

    @pytest.mark.parametrize(["network", "ip_version", "expected"], [
        ["0.0.0.0/0", 4, True],
        ["192.168.0.0/0", 4, False],
        ["::0/0", 6, True],
        ["2001:db00::0/24", 6, False],
    ])
    def test_normal(self, network, ip_version, expected):
        assert is_anywhere_network(network, ip_version) == expected

    @pytest.mark.parametrize(["network", "ip_version",  "expected"], [
        [None, 4, ValueError],
        ["0.0.0.0/0", None, ValueError],
    ])
    def test_exception(self, network, ip_version, expected):
        with pytest.raises(expected):
            is_anywhere_network(network, ip_version)


class Test_get_anywhere_network(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [4, "0.0.0.0/0"],
        ["4", "0.0.0.0/0"],
        [6, "::0/0"],
        ["6", "::0/0"],
    ])
    def test_normal(self, value, expected):
        assert get_anywhere_network(value) == expected

    @pytest.mark.parametrize(["value", "expected"], [
        [None, ValueError],
        ["test", ValueError],
    ])
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            get_anywhere_network(value)


class Test_sanitize_network(object):

    @pytest.mark.parametrize(["value", "ip_version", "expected"], [
        ["192.168.0.1", 4, "192.168.0.1/32"],
        ["192.168.0.1/32", 4, "192.168.0.1/32"],
        ["192.168.0.0/24", 4, "192.168.0.0/24"],
        ["192.168.0.0/23", 4, "192.168.0.0/23"],
        ["2001:db00::0/24", 6, "2001:db00::/24"],
        ["anywhere", 4, "0.0.0.0/0"],
        ["ANYWHERE", 4, "0.0.0.0/0"],
        ["anywhere", 6, "::0/0"],
        ["ANYWHERE", 6, "::0/0"],
        ["", 4, "0.0.0.0/0"],
        [None, 4, "0.0.0.0/0"],
        ["", 6, "::0/0"],
        [None, 6, "::0/0"],
    ])
    def test_normal(self, value, ip_version, expected):
        assert sanitize_network(value, ip_version) == expected

    @pytest.mark.parametrize(["value", "ip_version", "expected"], [
        ["192.168.0.1111", 4, ValueError],
        ["192.168.0.", 4, ValueError],
        [".168.0.1", 4, ValueError],
        ["192.168.1.0/23", 4, ValueError],
        ["192.168.0.0/33", 4, ValueError],
        ["192.168.1.0/24", 6, ValueError],
        ["2001:db00::0/24", 4, ValueError],
    ])
    def test_exception(self, value, ip_version, expected):
        with pytest.raises(expected):
            sanitize_network(value, ip_version)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import unicode_literals

import re

import pytest

from tcconfig._split_line_list import split_line_list


class Test_split_line_list(object):

    @pytest.mark.parametrize(
        [
            "value", "separator", "is_include_matched_line",
            "is_strip", "expected",
        ],
        [
            [
                [
                    "abcdefg",
                    "ABCDEFG",
                    "1234",
                ],
                re.compile("DEFG$"),
                False,
                True,
                [
                    ["abcdefg"],
                    ["1234"],
                ],
            ],
            [
                [
                    "abcdefg",
                    "ABCDEFG",
                    "ABCDEFG",
                    "1234",
                ],
                re.compile("DEFG$"),
                False,
                True,
                [
                    ["abcdefg"],
                    ["1234"],
                ],
            ],
            [
                [
                    "ABCDEFG",
                    "abcdefg",
                    "ABCDEFG",
                    "1234",
                    "ABCDEFG",
                ],
                re.compile("DEFG$"),
                False,
                True,
                [
                    ["abcdefg"],
                    ["1234"],
                ],
            ],
            [
                [
                    "abcdefg",
                    "ABCDEFG",
                    "1234"
                ],
                re.compile("DEFG$"),
                True,
                True,
                [
                    ["abcdefg"],
                    [
                        "ABCDEFG",
                        "1234",
                    ],
                ],
            ],
            [
                ["a", "  ", "b", "c"],
                re.compile("^$"),
                False,
                True,
                [
                    ["a"],
                    ["b", "c"]
                ],
            ],
            [
                ["a", "  ", "b", "c"],
                re.compile("^$"),
                False,
                False,
                [
                    ["a", "  ", "b", "c"],
                ],
            ],
        ])
    def test_normal(
            self, value, separator, is_include_matched_line, is_strip,
            expected):
        assert split_line_list(
            value, separator, is_include_matched_line, is_strip) == expected

    @pytest.mark.parametrize(
        [
            "value", "separator", "is_include_matched_line",
            "is_strip", "expected",
        ],
        [
            [None, "", False, True, TypeError],
            [["a", "b", "c"], None, False, True, AttributeError],
            [[1, 2, 3], re.compile(""), False, True, AttributeError],
            [[1, 2, 3], re.compile(""), False, False, TypeError],
        ])
    def test_exception(
            self, value, separator, is_include_matched_line, is_strip,
            expected):
        with pytest.raises(expected):
            split_line_list(
                value, separator, is_include_matched_line, is_strip)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import division
from __future__ import print_function

import json

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import TcCommand


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcconfig(object):

    @pytest.mark.xfail
    @pytest.mark.parametrize(["overwrite"], [
        [""],
        ["--overwrite"],
    ])
    def test_config_file_smoke(self, tmpdir, device_value, overwrite):
        if device_value is None:
            pytest.skip("device option is null")

        p = tmpdir.join("tcconfig.json")
        config = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
            "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                "delay": "10.0",
                "loss": "0.01",
                "rate": "250K",
                "delay-distro": "2.0"
            },
            "dst-network=0.0.0.0/0, src-port=1234, protocol=ip": {
                "delay": "50.0",
                "rate": "1G"
            }
        },
        "incoming": {
            "dst-network=192.168.10.0/24, protocol=ip": {
                "corrupt": "0.02",
                "rate": "1500K"
            }
        }
    }
}
"""
        p.write(config)

        device_option = "--device {:s}".format(device_value)

        SubprocessRunner("{:s} {:s}".format(
            TcCommand.TCDEL, device_option)).run()
        command = " ".join(
            ["{:s} -f ".format(TcCommand.TCSET), str(p), overwrite])
        SubprocessRunner(command).run()

        runner = SubprocessRunner("{:s} {:s}".format(
            TcCommand.TCSHOW, device_option))
        runner.run()

        print("[expected]\n{}\n".format(config))
        print("[actual]\n{}\n".format(runner.stdout))
        print("[stderr]\n{}".format(runner.stderr))

        assert json.loads(runner.stdout) == json.loads(config)

        SubprocessRunner("{:s} {:s}".format(
            TcCommand.TCDEL, device_option)).run()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

import itertools

import pytest

from allpairspy import AllPairs
from tcconfig._const import (
    Tc,
    TrafficDirection,
)
from tcconfig._error import (
    InvalidParameterError,
    UnitNotFoundError,
)
from tcconfig.traffic_control import TrafficControl

from .common import is_invalid_param


MIN_VALID_PACKET_LOSS = 0.0000000232  # [%]


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.mark.parametrize(["value"], [
    ["".join(opt_list)]
    for opt_list in AllPairs([
        ["0.1", "+1.25", "30"],
        [
            "k", " k ", "K", " K ", "kbps", "Kbps",
            "m", " m ", "M", " M ", "mbps", "Mbps",
            "g", " g ", "G", " G ", "gbps", "Gbps",
        ]
    ])
])
def test_TrafficControl_validate_bandwidth_rate_normal(value):
    tc_obj = TrafficControl(
        "dummy", bandwidth_rate=value, direction=TrafficDirection.OUTGOING)
    tc_obj.validate_bandwidth_rate()


@pytest.mark.parametrize(["value", "expected"], [
    ["".join(opt_list), UnitNotFoundError]
    for opt_list in AllPairs([
        ["0.1", "1"],
        ["", "kb", "KB", "mb", "MB", "gb", "GB"]
    ])
] + [
    ["".join(value), InvalidParameterError]
    for value in ("B", "K", "M", "G")
] + [
    ["0bps", InvalidParameterError],
    ["34359738361bps", InvalidParameterError],
])
def test_TrafficControl_validate_bandwidth_rate_exception_1(value, expected):
    with pytest.raises(expected):
        tc_obj = TrafficControl(
            "dummy", bandwidth_rate=value, direction=TrafficDirection.OUTGOING)
        tc_obj.validate_bandwidth_rate()


@pytest.mark.parametrize(["value", "expected"], [
    ["".join(opt_list), InvalidParameterError]
    for opt_list in itertools.product(
        ["-1", "0", "0.0"],
        ["k", "K", "m", "M", "g", "G"]
    )
])
def test_TrafficControl_validate_bandwidth_rate_exception_2(value, expected):
    with pytest.raises(expected):
        tc_obj = TrafficControl("dummy", bandwidth_rate=value)
        tc_obj.validate_bandwidth_rate()


class Test_TrafficControl_validate(object):

    @pytest.mark.parametrize(
        [
            "rate", "direction", "delay", "delay_distro", "loss", "duplicate",
            "corrupt", "network", "src_port", "dst_port",
        ],
        [
            opt_list
            for opt_list in AllPairs([
                [None, "", "100K", "0.5M", "0.1G"],  # rate
                [TrafficDirection.OUTGOING],
                [
                    None, TrafficControl.MIN_LATENCY_MS,
                    TrafficControl.MAX_LATENCY_MS,
                ],  # delay
                [
                    None, TrafficControl.MIN_LATENCY_MS,
                    TrafficControl.MAX_LATENCY_MS,
                ],  # delay_distro
                [
                    None,
                    TrafficControl.MIN_PACKET_LOSS_RATE, MIN_VALID_PACKET_LOSS,
                    TrafficControl.MAX_PACKET_LOSS_RATE,
                ],  # loss
                [
                    None, TrafficControl.MIN_PACKET_DUPLICATE_RATE,
                    TrafficControl.MAX_PACKET_DUPLICATE_RATE,
                ],  # duplicate
                [
                    None, TrafficControl.MIN_CORRUPTION_RATE,
                    TrafficControl.MAX_CORRUPTION_RATE,
                ],  # corrupt
                [
                    None,
                    "",
                    "192.168.0.1", "192.168.0.254",
                    "192.168.0.1/32", "192.168.0.0/24"
                ],  # network
                [None, 65535],  # src_port
                [None, 65535],  # dst_port
            ], n=3)
        ])
    def test_normal(
            self, device_option, rate, direction, delay, delay_distro,
            loss, duplicate, corrupt, network, src_port, dst_port):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            direction=direction,
            bandwidth_rate=rate,
            latency_ms=delay,
            latency_distro_ms=delay_distro,
            packet_loss_rate=loss,
            packet_duplicate_rate=duplicate,
            corruption_rate=corrupt,
            dst_network=network,
            src_port=src_port,
            dst_port=dst_port,
            src_network=None,
            is_enable_iptables=True,
        )

        if is_invalid_param(
                rate, delay, loss, duplicate, corrupt, reordering=None):
            with pytest.raises(ValueError):
                tc_obj.validate()
        else:
            tc_obj.validate()

    @pytest.mark.parametrize(
        ["direction", "delay", "reordering"],
        [
            opt_list
            for opt_list in AllPairs([
                [TrafficDirection.OUTGOING],
                [
                    TrafficControl.MIN_LATENCY_MS + 0.1,
                    TrafficControl.MAX_LATENCY_MS,
                ],  # delay
                [
                    TrafficControl.MIN_REORDERING_RATE,
                    TrafficControl.MAX_REORDERING_RATE,
                ],  # reordering
            ], n=2)
        ])
    def test_normal_reordering(
            self, device_option, direction, delay, reordering):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            direction=direction,
            latency_ms=delay,
            reordering_rate=reordering,
        )

        tc_obj.validate()

    @pytest.mark.parametrize(["value", "expected"], [
        [{"latency_ms": TrafficControl.MIN_LATENCY_MS - 1}, ValueError],
        [{"latency_ms": TrafficControl.MAX_LATENCY_MS + 1}, ValueError],

        [
            {
                "latency_ms": 100,
                "latency_distro_ms": TrafficControl.MIN_LATENCY_MS - 1,
            },
            ValueError
        ],
        [
            {
                "latency_ms": 100,
                "latency_distro_ms": TrafficControl.MAX_LATENCY_MS + 1,
            },
            ValueError
        ],

        [
            {"packet_loss_rate": TrafficControl.MIN_PACKET_LOSS_RATE - 0.1},
            ValueError
        ],
        [
            {"packet_loss_rate": TrafficControl.MAX_PACKET_LOSS_RATE + 0.1},
            ValueError
        ],

        [
            {
                "latency_ms": 100,
                "packet_duplicate_rate": TrafficControl.MIN_PACKET_DUPLICATE_RATE - 0.1,
            },
            ValueError
        ],
        [
            {
                "latency_ms": 100,
                "packet_duplicate_rate": TrafficControl.MAX_PACKET_DUPLICATE_RATE + 0.1,
            },
            ValueError
        ],

        [
            {"corruption_rate": TrafficControl.MIN_CORRUPTION_RATE - 0.1},
            ValueError
        ],
        [
            {"corruption_rate": TrafficControl.MAX_CORRUPTION_RATE + 0.1},
            ValueError
        ],

        [
            {"reordering_rate": TrafficControl.MIN_REORDERING_RATE - 0.1},
            ValueError
        ],
        [
            {"reordering_rate": TrafficControl.MAX_REORDERING_RATE + 0.1},
            ValueError
        ],


        [{Tc.Param.DST_NETWORK: "192.168.0."}, ValueError],
        [{Tc.Param.DST_NETWORK: "192.168.0.256"}, ValueError],
        [{Tc.Param.DST_NETWORK: "192.168.0.0/0"}, ValueError],
        [{Tc.Param.DST_NETWORK: "192.168.0.0/33"}, ValueError],
        [{Tc.Param.DST_NETWORK: "192.168.0.2/24"}, ValueError],
        [{Tc.Param.DST_NETWORK: "192.168.0.0000/24"}, ValueError],

        [{"src_port": -1}, ValueError],
        [{"src_port": 65536}, ValueError],

        [{"dst_port": -1}, ValueError],
        [{"dst_port": 65536}, ValueError],
    ])
    def test_exception(self, device_option, value, expected):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            bandwidth_rate=value.get("bandwidth_rate"),
            latency_ms=value.get("latency_ms"),
            latency_distro_ms=value.get("latency_distro_ms"),
            packet_loss_rate=value.get("packet_loss_rate"),
            packet_duplicate_rate=value.get("packet_duplicate_rate"),
            corruption_rate=value.get("corruption_rate"),
            dst_network=value.get(Tc.Param.DST_NETWORK),
            src_port=value.get("src_port"),
            dst_port=value.get("dst_port"),
        )

        with pytest.raises(expected):
            tc_obj.validate()


class Test_TrafficControl_ipv4(object):

    @pytest.mark.parametrize(
        [
            "network", "is_ipv6",
            "expected_ip_ver", "expected_protocol", "expected_protocol_match"
        ],
        [
            ["192.168.0.1", False, 4, "ip", "ip"],
            ["192.168.0.0/24", False, 4, "ip", "ip"],
            ["::1", True, 6, "ipv6", "ip6"],
            ["2001:db00::0/24", True, 6, "ipv6", "ip6"],
        ])
    def test_normal(
            self, device_option, network, is_ipv6,
            expected_ip_ver, expected_protocol, expected_protocol_match):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option, dst_network=network, is_ipv6=is_ipv6)

        assert tc_obj.ip_version == expected_ip_ver
        assert tc_obj.protocol == expected_protocol
        assert tc_obj.protocol_match == expected_protocol_match



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import unicode_literals

import os.path
import sys

import setuptools

import tcconfig


REQUIREMENT_DIR = "requirements"

needs_pytest = set(["pytest", "test", "ptr"]).intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []


with open("README.rst") as fp:
    long_description = fp.read()

with open(os.path.join("docs", "pages", "introduction", "summary.txt")) as f:
    summary = f.read()

with open(os.path.join(REQUIREMENT_DIR, "requirements.txt")) as f:
    install_requires = [line.strip() for line in f if line.strip()]

if any([
    sys.version_info.major < 3,
    sys.version_info.major == 3 and sys.version_info.minor < 3,
]):
    install_requires.append("ipaddress")

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    tests_require = [line.strip() for line in f if line.strip()]

setuptools.setup(
    name="tcconfig",
    version=tcconfig.VERSION,
    author="Tsuyoshi Hombashi",
    author_email="gogogo.vm@gmail.com",
    url="https://github.com/thombashi/tcconfig",
    description=summary,
    keywords=[
        "traffic control", "tc", "traffic shaping", "bandwidth",
        "latency", "packet loss",
    ],
    long_description=long_description,
    license="MIT License",
    include_package_data=True,
    packages=setuptools.find_packages(exclude=['test*']),
    install_requires=install_requires,
    setup_requires=pytest_runner,
    tests_require=tests_require,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: System :: Networking",
        "Topic :: System :: Systems Administration",
    ],
    entry_points={
        "console_scripts": [
            "tcset=tcconfig.tcset:main",
            "tcdel=tcconfig.tcdel:main",
            "tcshow=tcconfig.tcshow:main",
        ],
    }
)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import re


def __null_line_strip(line):
    return line


def __line_strip(line):
    return line.strip()


def split_line_list(
        line_list, re_line_separator=re.compile("^$"),
        is_include_match_line=False, is_strip=True):
    block_list = []
    block = []
    strip_func = __line_strip if is_strip else __null_line_strip

    for line in line_list:
        line = strip_func(line)

        if re_line_separator.search(line):
            if len(block) > 0:
                block_list.append(block)

            block = []
            if is_include_match_line:
                block.append(line)
            continue

        block.append(line)

    if len(block) > 0:
        block_list.append(block)

    return block_list



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division

import pingparsing
import pytest
from subprocrunner import SubprocessRunner
import typepy

from tcconfig._const import TcCommand

from .common import execute_tcdel


WAIT_TIME = 5  # [sec]


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def dst_host_ex_option(request):
    return request.config.getoption("--dst-host-ex")


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-f -q"
    transmitter.waittime = WAIT_TIME

    return transmitter


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


class Test_tcset_two_network(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:
      python setup.py test --addopts \
        "--device=<test device> --dst-host=<hostname/IP-addr> --dst-host-ex=<hostname/IP-addr>"

    These tests are expected to execute on following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    def test_network(
            self, device_option, dst_host_option, dst_host_ex_option,
            transmitter, pingparser):
        if device_option is None:
            pytest.skip("device option is null")

        if any([
            typepy.is_null_string(dst_host_option),
            typepy.is_null_string(dst_host_ex_option),
        ]):
            pytest.skip("destination host is null")

        execute_tcdel(device_option)
        delay = 100

        # tc to specific network ---
        command_list = [
            TcCommand.TCSET,
            "--device " + device_option,
            "--delay {:d}".format(delay),
            "--network " + dst_host_ex_option,
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        # w/o tc network ---
        transmitter.destination_host = dst_host_option
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/ tc network ---
        transmitter.destination_host = dst_host_ex_option
        result = transmitter.ping()
        pingparser.parse(result.stdout)
        with_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        # finalize ---
        execute_tcdel(device_option)



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
import argparse

import logbook

from ._const import TcCommandOutput


class ArgparseWrapper(object):
    """
    wrapper class of argparse
    """

    def __init__(self, version, description=""):
        epilog = "Issue tracker: https://github.com/thombashi/tcconfig/issues"

        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description, epilog=epilog)
        self.parser.add_argument(
            "--version", action="version", version="%(prog)s " + version)

        self._add_tc_command_arg_group()
        self._add_log_level_argument_group()

    def _add_log_level_argument_group(self):
        dest = "log_level"

        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--debug", dest=dest, action="store_const",
            const=logbook.DEBUG, default=logbook.INFO,
            help="for debug print.")
        group.add_argument(
            "--quiet", dest=dest, action="store_const",
            const=logbook.NOTSET, default=logbook.INFO,
            help="suppress execution log messages.")

        return group

    def _add_tc_command_arg_group(self):
        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--tc-command", dest="tc_command_output", action="store_const",
            const=TcCommandOutput.STDOUT, default=TcCommandOutput.NOT_SET,
            help="""
            display tc commands to be executed and exit.
            these commands are not actually executed.
            """)

        group.add_argument(
            "--tc-script", dest="tc_command_output", action="store_const",
            const=TcCommandOutput.SCRIPT, default=TcCommandOutput.NOT_SET,
            help="""
            generate a script file that described tc commands which equivalent
            with execution tcconfig command. the script can be execute without
            tcconfig package installation.
            """)
